#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of Pandora plugin LibSGM
#
#     https://github.com/CNES/Pandora_plugin_libsgm
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module provides class and functions to compute penalties used to optimize the cost volume using the LibSGM library
"""

from typing import Dict, Union, Tuple, List

import numpy as np
import xarray as xr
from json_checker import Checker, And, Or
from pandora.common import is_method

from pandora_plugin_libsgm.penalty import penalty


@penalty.AbstractPenalty.register_subclass("sgm_penalty")
class SgmPenalty(penalty.AbstractPenalty):
    """

    SGM Penalty

    """

    # Default configuration, do not change these values
    _P1 = 8
    _P2 = 32
    _ALPHA = 1.0
    _BETA = 1
    _GAMMA = 1
    _P2_METHOD = "constant"
    _OVERCOUNTING = False
    _MIN_COST_PATH = False
    _USE_CONFIDENCE = False
    _PIECEWISE_OPTIMIZATION_LAYER = "None"

    def __init__(self, directions: List[List[int]], **cfg: Union[str, int, float, bool]):
        """
        :param directions: directions to
        :type directions: list of [x offset, y offset]
        :param cfg: optional configuration, {'P1': value, 'P2': value, 'alpha': value, 'beta': value, 'gamma": value,
                                            'p2_method': value}
        :type cfg: dict
        """
        self.cfg = self.check_conf(**cfg)
        self._p1 = self.cfg["P1"]
        self._p2 = self.cfg["P2"]
        self._alpha = self.cfg["alpha"]
        self._beta = self.cfg["beta"]
        self._gamma = self.cfg["gamma"]
        self._overcounting = self.cfg["overcounting"]
        self._p2_method = self.cfg["p2_method"]
        self._min_cost_paths = self.cfg["min_cost_paths"]
        self._use_confidence = self.cfg["use_confidence"]
        self._piecewise_optimization_layer = self.cfg["piecewise_optimization_layer"]
        self._directions = directions

    def check_conf(self, **cfg: Union[str, int, float, bool]) -> Dict[str, Union[str, int, float, bool]]:
        """
        Add default values to the dictionary if there are missing elements and check if the dictionary is correct

        :param cfg: optimization configuration
        :type cfg: dict
        :return cfg: optimization configuration updated
        :rtype cfg: dict
        """
        # Give the default value if the required element is not in the configuration
        if "P1" not in cfg:
            cfg["P1"] = self._P1
        if "P2" not in cfg:
            cfg["P2"] = self._P2
        if "alpha" not in cfg:
            cfg["alpha"] = self._ALPHA
        if "beta" not in cfg:
            cfg["beta"] = self._BETA
        if "gamma" not in cfg:
            cfg["gamma"] = self._GAMMA
        if "p2_method" not in cfg:
            cfg["p2_method"] = self._P2_METHOD
        if "overcounting" not in cfg:
            cfg["overcounting"] = self._OVERCOUNTING
        if "min_cost_paths" not in cfg:
            cfg["min_cost_paths"] = self._MIN_COST_PATH
        if "use_confidence" not in cfg:
            cfg["use_confidence"] = self._USE_CONFIDENCE
        if "piecewise_optimization_layer" not in cfg:
            cfg["piecewise_optimization_layer"] = self._PIECEWISE_OPTIMIZATION_LAYER

        p1_value = cfg["P1"]

        schema = {
            "sgm_version": And(str, lambda x: is_method(x, ["c++", "python_libsgm", "python_libsgm_parall"])),
            "optimization_method": And(str, lambda x: is_method(x, ["sgm"])),
            "penalty_method": And(str, lambda x: is_method(x, ["sgm_penalty"])),
            "P1": And(Or(int, float), lambda x: x > 0),
            "P2": And(Or(int, float), lambda x: x > p1_value),
            "alpha": And(Or(int, float), lambda x: x >= 0),
            "beta": And(Or(int, float), lambda x: x > 0),
            "gamma": And(Or(int, float), lambda x: x > 0),
            "p2_method": And(str, lambda x: is_method(x, ["constant", "negativeGradient", "inverseGradient"])),
            "overcounting": bool,
            "min_cost_paths": bool,
            "use_confidence": bool,
            "piecewise_optimization_layer": And(str, lambda x: is_method(x, ["None", "classif", "segm"])),
        }

        checker = Checker(schema)
        checker.validate(cfg)
        return cfg

    def desc(self):
        """
        Describes the penality method

        """
        print("SGM penalty method description")

    def compute_penalty(
        self, cv: xr.Dataset, img_left: xr.Dataset, img_right: xr.Dataset
    ) -> Tuple[float, np.ndarray, np.ndarray]:
        """
        Compute penalty

        :param cv: the cost volume, with the data variables:

            - cost_volume 3D xarray.DataArray (row, col, disp)
            - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv: xarray.Dataset
        :param img_left: left  image
        :type img_left: xarray.Dataset
        :param img_right: right  image
        :type img_right: xarray.Dataset
        :return: P1 and P2 penalties
        :rtype: tuple(numpy array, numpy array)
        """

        # Get array
        img_left_array = img_left["im"].data

        # Calculation of the invalid value according to the chosen P2 estimation method
        invalid_value = None
        if self._p2_method == "constant":
            invalid_value = float(cv.attrs["cmax"] + self._p2 + 1)
        elif self._p2_method == "negativeGradient":
            invalid_value = float(cv.attrs["cmax"] + self._gamma + 1)
        elif self._p2_method == "inverseGradient":
            invalid_value = float(cv.attrs["cmax"] + self._gamma + (self._alpha / self._beta) + 1)  # type: ignore

        # Compute penalties
        if self._p2_method == "negativeGradient":
            p1_mask, p2_mask = self.negative_penalty_function(
                img_left_array,
                self._p1,  # type: ignore
                self._p2,  # type: ignore
                self._directions,
                self._alpha,  # type: ignore
                self._gamma,  # type: ignore
            )

        elif self._p2_method == "inverseGradient":
            p1_mask, p2_mask = self.inverse_penalty_function(
                img_left_array,
                self._p1,  # type: ignore
                self._p2,  # type: ignore
                self._directions,
                self._alpha,  # type: ignore
                self._beta,  # type: ignore
                self._gamma,  # type: ignore
            )

        else:
            # Default p2_method is constant
            p1_mask, p2_mask = self.constant_penalty_function(
                img_left_array, self._p1, self._p2, self._directions  # type: ignore
            )

        return invalid_value, p1_mask, p2_mask

    def negative_penalty_function(
        self,
        img_left: np.ndarray,
        p1: Union[int, float],
        p2: Union[int, float],
        directions: List[List[int]],
        alpha: float,
        gamma: float,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute negative penalty

        :param img_left: left  image
        :type img_left: numpy array
        :param p1: default P1 penalty
        :type p1: int or float
        :param p2: default P2 penalty
        :type p2: int or float
        :param directions: directions to
        :type directions: list of [x offset, y offset]
        :param alpha: hyper parameter
        :type alpha: float
        :param gamma: hyper parameter
        :type gamma: float
        :return: P1 and P2 penalties
        :rtype: tuple(numpy array, numpy array)
        """
        p1_mask = p1 * np.ones([img_left.shape[0], img_left.shape[1], len(directions)], dtype=np.float32)
        p2_mask = p2 * np.ones([img_left.shape[0], img_left.shape[1], len(directions)], dtype=np.float32)

        nb_directions = len(directions)
        for i in range(nb_directions):
            direction = directions[i]
            abs_gradient = self.compute_gradient(img_left[:, :], direction)
            val = -alpha * abs_gradient + gamma
            p2_mask[
                max(0, direction[0]) : min(img_left.shape[0] + direction[0], img_left.shape[0]),
                max(0, direction[1]) : min(img_left.shape[1] + direction[1], img_left.shape[1]),
                i,
            ] = val
        # if p2 < defaultP2 then p2
        msk = p2_mask < p2
        p2_mask = p2 * msk + p2_mask * (1 - msk)
        return p1_mask, p2_mask

    def inverse_penalty_function(
        self,
        img_left: np.ndarray,
        p1: Union[int, float],
        p2: Union[int, float],
        directions: List[List[int]],
        alpha: float,
        beta: float,
        gamma: float,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute inverse penalty

        :param img_left: left  image
        :type img_left: numpy array
        :param p1: P1 penalty
        :type p1: int or float
        :param p2: P2 penalty
        :type p2: int or float
        :param directions: directions to
        :type directions: list of [x offset, y offset]
        :param alpha: hyper parameter
        :type alpha: float
        :param beta: hyper parameter
        :type beta: float
        :param gamma: hyper parameter
        :type gamma: float
        :return: P1 and P2 penalties
        :rtype: tuple(numpy array, numpy array)
        """
        p1_mask = p1 * np.ones([img_left.shape[0], img_left.shape[1], len(directions)], dtype=np.float32)
        p2_mask = p2 * np.ones([img_left.shape[0], img_left.shape[1], len(directions)], dtype=np.float32)

        nb_directions = len(directions)
        for i in range(nb_directions):
            direc = directions[i]
            abs_gradient = self.compute_gradient(img_left[:, :], direc)
            val = alpha / (abs_gradient + beta) + gamma
            p2_mask[
                max(0, direc[0]) : min(img_left.shape[0] + direc[0], img_left.shape[0]),
                max(0, direc[1]) : min(img_left.shape[1] + direc[1], img_left.shape[1]),
                i,
            ] = val
        # if p2 < defaultP2 then p2
        msk = p2_mask < p2
        p2_mask = p2 * msk + p2_mask * (1 - msk)
        return p1_mask, p2_mask

    @staticmethod
    def constant_penalty_function(
        img_left: np.ndarray, p1: Union[int, float], p2: Union[int, float], directions: List[List[int]]
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute constant penalty

        :param img_left: left  image
        :type img_left: numpy array
        :param p1: P1 penalty
        :type p1: int or float
        :param p2: P2 penalty
        :type p2: int or float
        :param directions: directions to
        :type directions: list of [x offset, y offset]
        :return: P1 and P2 penalties
        :rtype: tuple(numpy array, numpy array)
        """
        p1_mask = p1 * np.ones([img_left.shape[0], img_left.shape[1], len(directions)], dtype=np.float32)
        p2_mask = p2 * np.ones([img_left.shape[0], img_left.shape[1], len(directions)], dtype=np.float32)
        return p1_mask, p2_mask



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains common functions present in Plugin_libsgm's tests.
"""


def error_mask(data, gt):
    """
    Percentage of bad pixels ( != ground truth ) in the validity mask

    """
    nb_rows, nb_cols = data.shape
    nb_error = 0
    for row in range(nb_rows):
        for col in range(nb_cols):
            if data[row, col] != gt[row, col]:
                nb_error += 1

    return nb_error / float(nb_rows * nb_cols)


def strict_error(data, gt):
    """
    Average of bad pixels  ( != ground truth )

    """
    nb_rows, nb_cols = data.shape
    nb_error = 0
    for row in range(nb_rows):
        for col in range(nb_cols):
            if data[row, col] != gt[row, col]:
                nb_error += 1

    return nb_error / float(nb_rows * nb_cols)


def error(data, gt, threshold, unknown_disparity=0):
    """
    Percentage of bad pixels whose error is > threshold

    """
    nb_row, nb_col = data.shape
    nb_error = 0
    for row in range(nb_row):
        for col in range(nb_col):
            if gt[row, col] != unknown_disparity:
                if abs((data[row, col] + gt[row, col])) > threshold:
                    nb_error += 1

    return nb_error / float(nb_row * nb_col)



# pylint: skip-file
#
# coding: utf8
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of Pandora plugin LibSGM
#
#     https://github.com/CNES/Pandora_plugin_libsgm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = "Pandora Plugin Libsgm"
copyright = "2020, CNES"
author = "CNES"
version = "0.6.0"
# The full version, including alpha/beta/rc tags
# not useful at the moment :  see  https://gitlab.cnes.fr/OutilsCommuns/CorrelateurChaine3D/pandora/issues/124
# release = '0.2.0'
# version = '0.2.0'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc",
    "sphinx_rtd_theme",
    "sphinx.ext.imgmath",
    "autoapi.extension",
]

autoapi_dirs = ["../../pandora_plugin_libsgm"]
autoapi_root = "api_reference"
autoapi_keep_files = True
autoapi_options = [
    "members",
    "undoc-members",
    "private-members",
    "show-inheritance",
    "show-module-summary",
    "special-members",
]

# Add any paths that contain templates here, relative to this directory.cd
templates_path = ["_templates"]

# These folders are copied to the documentation's HTML output
html_static_path = ["_static"]

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = ["css/my_custom.css"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

html_theme_options = {
    "logo_only": True,
    "navigation_depth": 3,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

latex_elements = {"papersize": "letterpaper", "pointsize": "10pt", "preamble": "", "figure_align": "htbp"}

numfig = True



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of Pandora plugin LibSGM
#
#     https://github.com/CNES/Pandora_plugin_libsgm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module provides functions to test mc-cnn penalties
"""

import unittest

import numpy as np

from pandora_plugin_libsgm.penalty import penalty_mc_cnn


class TestPenalitySGM(unittest.TestCase):
    """
    TestPenalitySGM class allows to test penality_sgm
    """

    def setUp(self):
        """
        Method called to prepare the configuration

        """
        self.cfg = {
            "sgm_version": "c++",
            "optimization_method": "sgm",
            "P1": 8,
            "P2": 10,
            "Q1": 1.0,
            "Q2": 1.0,
            "D": 1.0,
            "V": 1.0,
            "overcounting": False,
            "min_cost_paths": False,
            "penalty_method": "mc_cnn_fast_penalty",
        }

        self._directions = [[0, 1], [1, 0], [1, 1], [1, -1], [0, -1], [-1, 0], [-1, -1], [-1, 1]]

        self.penalty = penalty_mc_cnn.MccnnPenalty(self._directions, **self.cfg)

    def test_gradient(self):
        """
        Test Computation of gradient

        """

        img_left = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        # TEST 1
        img_wanted = np.array([[3, 3, 3], [3, 3, 3]])

        direc = [1, 0]
        computed_gradient = self.penalty.compute_gradient(img_left, direc)
        # Check if the calculated gradient is equal to the ground truth (same shape and all elements equals)
        np.testing.assert_array_equal(computed_gradient, img_wanted)

        # TEST 2
        img_wanted = np.array([[1, 1], [1, 1], [1, 1]])

        direc = [0, 1]
        computed_gradient = self.penalty.compute_gradient(img_left, direc)
        # Check if the calculated gradient is equal to the ground truth (same shape and all elements equals)
        np.testing.assert_array_equal(computed_gradient, img_wanted)

        # TEST 3
        img_wanted = np.array([[4, 4], [4, 4]])

        direc = [1, 1]
        computed_gradient = self.penalty.compute_gradient(img_left, direc)
        # Check if the calculated gradient is equal to the ground truth (same shape and all elements equals)
        np.testing.assert_array_equal(computed_gradient, img_wanted)

        # TEST 4
        img_wanted = np.array([[4, 4], [4, 4]])

        direc = [-1, -1]
        computed_gradient = self.penalty.compute_gradient(img_left, direc)
        # Check if the calculated gradient is equal to the ground truth (same shape and all elements equals)
        np.testing.assert_array_equal(computed_gradient, img_wanted)

        # TEST 5
        img_wanted = np.array([[2, 2], [2, 2]])

        direc = [-1, 1]
        computed_gradient = self.penalty.compute_gradient(img_left, direc)
        # Check if the calculated gradient is equal to the ground truth (same shape and all elements equals)
        np.testing.assert_array_equal(computed_gradient, img_wanted)

    def test_mc_cnn_penalty_function(self):
        """
        Test mc_cnn menalty

        """

        # TEST 1

        default_p1 = 3
        default_p2 = 4
        d_mccnn = 5
        v_mccnn = 2
        q1_mccnn = 5
        q2_mccnn = 6

        img_right = np.array(
            [[1, 6, 11, 16, 21], [2, 7, 12, 17, 22], [3, 8, 13, 18, 23], [4, 9, 14, 19, 24], [5, 10, 15, 20, 25]]
        )

        img_left = np.array(
            [[1, 2, 3, 4, 5], [6, 7, 8, 9, 10], [11, 12, 13, 14, 15], [16, 17, 18, 19, 20], [21, 22, 23, 24, 25]]
        )

        p1_wanted_0 = np.array(
            [
                [3, 3, 3, 3, 3],
                [3 / 5, 3 / 5, 3 / 5, 3 / 5, 3 / 5],
                [3 / 5, 3 / 5, 3 / 5, 3 / 5, 3 / 5],
                [3 / 5, 3 / 5, 3 / 5, 3 / 5, 3 / 5],
                [3 / 5, 3 / 5, 3 / 5, 3 / 5, 3 / 5],
            ],
            dtype=np.float32,
        )

        p2_wanted_0 = np.array(
            [
                [4, 4, 4, 4, 4],
                [4 / 5, 4 / 5, 4 / 5, 4 / 5, 4 / 5],
                [4 / 5, 4 / 5, 4 / 5, 4 / 5, 4 / 5],
                [4 / 5, 4 / 5, 4 / 5, 4 / 5, 4 / 5],
                [4 / 5, 4 / 5, 4 / 5, 4 / 5, 4 / 5],
            ],
            dtype=np.float32,
        )

        # divide by V on direction in [1,5]
        p1_wanted_1 = np.array(
            [
                [3 / 2, 3 / 2, 3 / 2, 3 / 2, 3 / 2],
                [3 / 2, 3 / 2, 3 / 2, 3 / 2, 3 / 2],
                [3 / 2, 3 / 2, 3 / 2, 3 / 2, 3 / 2],
                [3 / 2, 3 / 2, 3 / 2, 3 / 2, 3 / 2],
                [3 / 2, 3 / 2, 3 / 2, 3 / 2, 3 / 2],
            ],
            dtype=np.float32,
        )

        p2_wanted_1 = np.array([[4, 4, 4, 4, 4], [4, 4, 4, 4, 4], [4, 4, 4, 4, 4], [4, 4, 4, 4, 4], [4, 4, 4, 4, 4]])

        p1_wanted_2 = np.array(
            [
                [3, 3, 3, 3, 3],
                [3, 3 / (5 * 6), 3 / (5 * 6), 3 / (5 * 6), 3 / (5 * 6)],
                [3, 3 / (5 * 6), 3 / (5 * 6), 3 / (5 * 6), 3 / (5 * 6)],
                [3, 3 / (5 * 6), 3 / (5 * 6), 3 / (5 * 6), 3 / (5 * 6)],
                [3, 3 / (5 * 6), 3 / (5 * 6), 3 / (5 * 6), 3 / (5 * 6)],
            ],
            dtype=np.float32,
        )

        p2_wanted_2 = np.array(
            [
                [4, 4, 4, 4, 4],
                [4, 4 / (5 * 6), 4 / (5 * 6), 4 / (5 * 6), 4 / (5 * 6)],
                [4, 4 / (5 * 6), 4 / (5 * 6), 4 / (5 * 6), 4 / (5 * 6)],
                [4, 4 / (5 * 6), 4 / (5 * 6), 4 / (5 * 6), 4 / (5 * 6)],
                [4, 4 / (5 * 6), 4 / (5 * 6), 4 / (5 * 6), 4 / (5 * 6)],
            ],
            dtype=np.float32,
        )

        directions = [[1, 0], [-1, 1], [1, 1]]
        computed_p1, computed_p2 = self.penalty.mc_cnn_penalty_function(
            img_left, img_right, default_p1, default_p2, q1_mccnn, q2_mccnn, d_mccnn, v_mccnn, directions
        )
        # Check if the calculated gradient is equal to the ground truth (same shape and all elements equals)
        np.testing.assert_array_equal(computed_p1[:, :, 0], p1_wanted_0)
        np.testing.assert_array_equal(computed_p2[:, :, 0], p2_wanted_0)
        np.testing.assert_array_equal(computed_p1[:, :, 1], p1_wanted_1)
        np.testing.assert_array_equal(computed_p2[:, :, 1], p2_wanted_1)
        np.testing.assert_array_equal(computed_p1[:, :, 2], p1_wanted_2)
        np.testing.assert_array_equal(computed_p2[:, :, 2], p2_wanted_2)

    # pylint:disable=protected-access
    @staticmethod
    def test_penalty_default_configuration():
        """
        Test default parameters according to the type of mc-cnn measure

        """
        # Check with fast mc-cnn
        cfg = {
            "sgm_version": "c++",
            "optimization_method": "sgm",
            "overcounting": False,
            "min_cost_paths": False,
            "penalty_method": "mc_cnn_fast_penalty",
        }

        _directions = [[0, 1], [1, 0], [1, 1], [1, -1], [0, -1], [-1, 0], [-1, -1], [-1, 1]]
        penalty = penalty_mc_cnn.MccnnPenalty(_directions, **cfg)

        # Checks that penalties correspond to the type of measure
        assert penalty._p1 == penalty_mc_cnn.MccnnPenalty._P1_FAST
        assert penalty._p2 == penalty_mc_cnn.MccnnPenalty._P2_FAST
        assert penalty._q1 == penalty_mc_cnn.MccnnPenalty._Q1_FAST
        assert penalty._q2 == penalty_mc_cnn.MccnnPenalty._Q2_FAST
        assert penalty._d == penalty_mc_cnn.MccnnPenalty._D_FAST
        assert penalty._v == penalty_mc_cnn.MccnnPenalty._V_FAST

        # Check with accurate mc-cnn
        cfg = {
            "sgm_version": "c++",
            "optimization_method": "sgm",
            "overcounting": False,
            "min_cost_paths": False,
            "penalty_method": "mc_cnn_accurate_penalty",
        }

        _directions = [[0, 1], [1, 0], [1, 1], [1, -1], [0, -1], [-1, 0], [-1, -1], [-1, 1]]
        penalty = penalty_mc_cnn.MccnnPenalty(_directions, **cfg)

        # Checks that penalties correspond to the type of measure
        assert penalty._p1 == penalty_mc_cnn.MccnnPenalty._P1_ACCURATE
        assert penalty._p2 == penalty_mc_cnn.MccnnPenalty._P2_ACCURATE
        assert penalty._q1 == penalty_mc_cnn.MccnnPenalty._Q1_ACCURATE
        assert penalty._q2 == penalty_mc_cnn.MccnnPenalty._Q2_ACCURATE
        assert penalty._d == penalty_mc_cnn.MccnnPenalty._D_ACCURATE
        assert penalty._v == penalty_mc_cnn.MccnnPenalty._V_ACCURATE
